/*
 * Copyright (c) 2014, National Institute of Informatics
 * Copyright (c) 2011, SWITCH
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of SWITCH nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL SWITCH BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package ch.SWITCH.aai.uApprove.ar;

import java.io.IOException;
import java.util.HashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;

import javax.servlet.ServletContext;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import jp.gakunin.uApprovejp.ar.LocalizedIntendedUseMap;
import jp.gakunin.uApprovejp.ar.PageControl;
import jp.gakunin.uApprovejp.ar.SelectableAttribute;
import jp.gakunin.uApprovejp.ar.UnknownControlException;
import jp.gakunin.uApprovejp.consent.ConsentMode;
import jp.gakunin.uApprovejp.consent.ConsentOperation;
import jp.gakunin.uApprovejp.consent.ConsentOperationFacrtory;
import jp.gakunin.uApprovejp.sad.ServiceAccessDataModule;

import org.opensaml.saml2.metadata.LocalizedString;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.web.context.WebApplicationContext;
import org.springframework.web.context.support.WebApplicationContextUtils;

import ch.SWITCH.aai.uApprove.IdPHelper;
import ch.SWITCH.aai.uApprove.ViewHelper;

/**
 * Attribute Release Servlet.
 */
public class AttributeReleaseServlet extends HttpServlet {

    /** Serial version UID. */
    private static final long serialVersionUID = 1L;

    /** Class logger. */
    private final Logger logger = LoggerFactory.getLogger(AttributeReleaseServlet.class);

    /** The servlet context. */
    private ServletContext servletContext;

    /** The Service Access Data Module */
    private ServiceAccessDataModule serviceAccessDataModule;

    /** The Attribute Release module. */
    private AttributeReleaseModule attributeReleaseModule;

    private ConsentOperationFacrtory consentOperationFactory;

    /** The SAML helper. */
    private SAMLHelper samlHelper;

    /** The view helper. */
    private ViewHelper viewHelper;

    /** {@inheritDoc} */
    public void init() throws ServletException {
        try {
            super.init();
            servletContext = getServletContext();
            final WebApplicationContext appContext =
                    WebApplicationContextUtils.getRequiredWebApplicationContext(servletContext);
            serviceAccessDataModule =
                    (ServiceAccessDataModule) appContext.getBean("uApprove.serviceAccessDataModule",
                            ServiceAccessDataModule.class);
            attributeReleaseModule =
                    (AttributeReleaseModule) appContext.getBean("uApprove.attributeReleaseModule",
                            AttributeReleaseModule.class);
            samlHelper = (SAMLHelper) appContext.getBean("uApprove.samlHelper", SAMLHelper.class);
            viewHelper = (ViewHelper) appContext.getBean("uApprove.viewHelper", ViewHelper.class);

            consentOperationFactory =
                    new ConsentOperationFacrtory(serviceAccessDataModule, attributeReleaseModule, samlHelper);
        } catch (final Throwable t) {
            logger.error("Error while initializing Attribute Release Servlet.", t);
            throw new ServletException(t);
        }
    }

    /** {@inheritDoc} */
    protected void doGet(final HttpServletRequest req, final HttpServletResponse resp) throws ServletException,
            IOException {
        try {
            final ConsentMode cm = IdPHelper.getConsentMode(servletContext, req);
            final String relyingPartyId = IdPHelper.getRelyingPartyId(servletContext, req);
            final LocalizedIntendedUseMap localizedIntendedUseMap =
                    IdPHelper.getLocalizedIntendedUseMap(servletContext, req);
            final List<SelectableAttribute> attributes = IdPHelper.getAttributes(servletContext, req);
            final List<SelectableAttribute> mandatoryAttributes = IdPHelper.getMandatoryAttributes(servletContext, req);

            final Map<String, Object> context = new HashMap<String, Object>();
            final Locale locale = viewHelper.selectLocale(req);

            registerLocalizedIntendedUseToAttribute(attributes, localizedIntendedUseMap, locale);
            registerLocalizedIntendedUseToAttribute(mandatoryAttributes, localizedIntendedUseMap, locale);

            context.put("relyingParty", samlHelper.readRelyingParty(relyingPartyId, locale));
            context.put("attributes", attributes);
            context.put("mandatoryAttributes", mandatoryAttributes);
            context.put(ConsentMode.Parameter, cm);
            context.put("allowGeneralConsent", attributeReleaseModule.isAllowGeneralConsent());
            viewHelper.showView(servletContext, req, resp, "attribute-release", context);
        } catch (final Throwable t) {
            logger.error("Error while GET Attribute Release Servlet.", t);
            IdPHelper.handleException(servletContext, req, resp, t);
        }
    }

    /** {@inheritDoc} */
    protected void doPost(final HttpServletRequest req, final HttpServletResponse resp) throws ServletException,
            IOException {
        try {
            final ConsentMode cm = ConsentMode.getConsentMode(req);
            logger.debug("ConsentMode = {}", cm);
            final List<SelectableAttribute> attributes = IdPHelper.getAttributes(servletContext, req);
            final List<SelectableAttribute> mandatoryAttributes = IdPHelper.getMandatoryAttributes(servletContext, req);

            // "Next" from attribute-release.jsp
            if (PageControl.NEXT.isRequested(req)) {
                logger.debug("coming from attribute-release.jsp Next, redirect to attribute-check page");
                updateSelectionStatus(req, attributes);

                final Map<String, Object> context = new HashMap<String, Object>();
                context.put("attributes", attributes);
                context.put("mandatoryAttributes", mandatoryAttributes);
                context.put(ConsentMode.Parameter, cm);
                viewHelper.showView(servletContext, req, resp, "attribute-check", context);
                return;
            }

            // "Back" from attribute-check.jsp
            if (PageControl.BACK.isRequested(req)) {
                logger.debug("coming from attribute-check back, redirect to attribute-release page");
                // TODO update descriptions
                final LocalizedIntendedUseMap localizedIntendedUseMap =
                        IdPHelper.getLocalizedIntendedUseMap(servletContext, req);
                final Locale locale = viewHelper.selectLocale(req);

                registerLocalizedIntendedUseToAttribute(attributes, localizedIntendedUseMap, locale);
                registerLocalizedIntendedUseToAttribute(mandatoryAttributes, localizedIntendedUseMap, locale);

                final Map<String, Object> context = new HashMap<String, Object>();
                context.put("attributes", attributes);
                context.put("mandatoryAttributes", mandatoryAttributes);
                context.put(ConsentMode.Parameter, cm);
                context.put("allowGeneralConsent", attributeReleaseModule.isAllowGeneralConsent());
                viewHelper.showView(servletContext, req, resp, "attribute-release", context);
                return;
            }

            // "Send" from attribute-check.jsp
            if (PageControl.SEND.isRequested(req)) {
                logger.debug("coming from attribute-check send.");
                req.setAttribute("attributes", attributes);
                req.setAttribute("mandatoryAttributes", mandatoryAttributes);

                final ConsentOperation consentOperation = cm.getOperation(consentOperationFactory);
                consentOperation.consent(servletContext, req);

                IdPHelper.setAttributeReleaseConsented(servletContext, req);
                IdPHelper.returnToIdP(servletContext, req, resp);
                return;
            }

            // coming here is not ok
            throw new UnknownControlException("uApprove viewer flow error");
        } catch (final Throwable t) {
            logger.error("Error while POST Attribute Release Servlet.", t);
            IdPHelper.handleException(servletContext, req, resp, t);
        }
    }

    /**
     * Register the localized description about intended uses to atteibutes.
     * 
     * @param attributes
     * @param map
     * @param locale
     */
    private void registerLocalizedIntendedUseToAttribute(final List<SelectableAttribute> attributes,
            final LocalizedIntendedUseMap map, final Locale locale) {
        for (final SelectableAttribute sAttr : attributes) {
            if (map.containsKey(sAttr.getId(), locale)) {
                final LocalizedString intendesUse = map.getIntendesUse(sAttr.getId(), locale);
                sAttr.setLocalizedIntendedUse(intendesUse.getLocalString());
            } else {
                sAttr.setLocalizedIntendedUse(null);
            }
        }
    }

    /**
     * Updates the selection status of each attributes.
     * 
     * Updates the selection status of each attributes by using parameter of posted parameter.
     * 
     * @param req
     * @param attributes
     */
    private void updateSelectionStatus(final HttpServletRequest req, final List<SelectableAttribute> attributes) {
        final String[] selectedAttributeIds = req.getParameterValues("approved-attributes[]");

        if (selectedAttributeIds == null) {
            return;
        }

        for (final SelectableAttribute attribute : attributes) {
            attribute.setSelected(false);
            for (final String selectedId : selectedAttributeIds) {
                if (selectedId.equals(attribute.getId())) {
                    logger.debug("{} was selected.", attribute.getId());
                    attribute.setSelected(true);
                    break;
                }
            }
            if (!attribute.isSelected()) {
                logger.debug("{} was not consented.", attribute.getId());
            }
        }

        IdPHelper.setAttributes(servletContext, req, attributes);
    }
}
