/*
 * Copyright (c) 2014, National Institute of Informatics
 */

package jp.gakunin.shibboleth.common.attribute.filtering.provider.match.uapprove;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import jp.gakunin.uApprovejp.ar.UApprovejpSAMLHelper;
import jp.gakunin.uApprovejp.ar.UApproveBaseSAMLProfileRequestContext;
import jp.gakunin.uApprovejp.consent.ConsentMode;
import jp.gakunin.uApprovejp.sad.ServiceAccessDataModule;

import org.joda.time.DateTime;
import org.opensaml.common.SAMLObject;
import org.opensaml.saml2.core.AuthnRequest;
import org.opensaml.saml2.metadata.AttributeConsumingService;
import org.opensaml.saml2.metadata.RequestedAttribute;
import org.opensaml.saml2.metadata.RoleDescriptor;
import org.opensaml.saml2.metadata.SPSSODescriptor;
import org.opensaml.xml.XMLObject;
import org.opensaml.xml.schema.XSAny;
import org.opensaml.xml.schema.XSBase64Binary;
import org.opensaml.xml.schema.XSBoolean;
import org.opensaml.xml.schema.XSDateTime;
import org.opensaml.xml.schema.XSInteger;
import org.opensaml.xml.schema.XSString;
import org.opensaml.xml.schema.XSURI;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import ch.SWITCH.aai.uApprove.ar.AttributeReleaseConsent;
import ch.SWITCH.aai.uApprove.ar.AttributeReleaseModule;
import edu.internet2.middleware.shibboleth.common.attribute.AttributeRequestContext;
import edu.internet2.middleware.shibboleth.common.attribute.BaseAttribute;
import edu.internet2.middleware.shibboleth.common.attribute.encoding.AttributeEncoder;
import edu.internet2.middleware.shibboleth.common.attribute.filtering.provider.FilterProcessingException;
import edu.internet2.middleware.shibboleth.common.attribute.filtering.provider.ShibbolethFilteringContext;
import edu.internet2.middleware.shibboleth.common.attribute.filtering.provider.match.basic.AbstractMatchFunctor;
import edu.internet2.middleware.shibboleth.common.attribute.provider.BasicAttribute;
import edu.internet2.middleware.shibboleth.common.profile.provider.SAMLProfileRequestContext;

/**
 * Match functor that checks whether an attribute is enumerated in an SP's metadata as a required or optional attribute.
 * 
 * <table border="1">
 * <tr>
 * <th>matchIfMetadataSilent</th>
 * <th>onlyIfChecked</th>
 * <th>result</th>
 * </tr>
 * <tr>
 * <td>false</td>
 * <td>N/A</td>
 * <td>true</td>
 * </tr>
 * <td rowspan="3">true</td>
 * <td rowspan="2">true</td>
 * <td>true</td>
 * </tr>
 * <tr>
 * <td>false (when mandatory mode)</td>
 * </tr>
 * <tr>
 * <td>false</td>
 * <td>false</td>
 * </tr>
 * </table>
 * 
 * 
 * <table border="1">
 * <tr>
 * <th>isRequired</th>
 * <th>onlyIfRequred</th>
 * <th>onlyIfChecked</th>
 * <th>result</th>
 * </tr>
 * <tr>
 * <td>true</td>
 * <td>true</td>
 * <td>N/A</td>
 * <td>true</td>
 * </tr>
 * <tr>
 * <td>true</td>
 * <td>false</td>
 * <td>N/A</td>
 * <td>true</td>
 * </tr>
 * <tr>
 * <td>false</td>
 * <td>true</td>
 * <td>N/A</td>
 * <td>false</td>
 * </tr>
 * <tr>
 * <td rowspan="3">false</td>
 * <td rowspan="3">false</td>
 * <td rowspan="2">true</td>
 * <td>true</td>
 * </tr>
 * <tr>
 * <td>false (when mandatory mode)</td>
 * </tr>
 * <tr>
 * <td>false</td>
 * <td>false</td>
 * </tr>
 * </table>
 */
public class AttributeInMetadataMatchFunctor extends AbstractMatchFunctor {

    /** Class logger. */
    private final Logger log = LoggerFactory.getLogger(AttributeInMetadataMatchFunctor.class);

    public static final String MANDATORY_MODE = "mandatory";

    /** Whether optionally requested attributes should be matched. */
    private boolean onlyIfRequired;

    /** Whether to return a match if the metadata does not contain an ACS descriptor. */
    private boolean matchIfMetadataSilent;

    /** Whether the attribute is approved. */
    private boolean onlyIfChecked;

    private ServiceAccessDataModule serviceAccessDataModule;

    /** bean */
    private AttributeReleaseModule attributeReleaseModule;

    /**
     * Gets whether optionally requested attributes should be matched.
     * 
     * @return Whether optionally requested attributes should be matched.
     */
    public boolean isOnlyIfRequired() {
        return onlyIfRequired;
    }

    /**
     * Sets whether optionally requested attributes should be matched.
     * 
     * @param flag whether optionally requested attributes should be matched
     */
    public void setOnlyIfRequired(final boolean flag) {
        onlyIfRequired = flag;
    }

    /**
     * Gets whether to matched if the metadata contains no AttributeConsumingService.
     * 
     * @return whether to match if the metadata contains no AttributeConsumingService
     */
    public boolean isMatchIfMetadataSilent() {
        return matchIfMetadataSilent;
    }

    /**
     * Sets whether to match if the metadata contains no AttributeConsumingService.
     * 
     * @param flag whether to match if the metadata contains no AttributeConsumingService
     */
    public void setMatchIfMetadataSilent(final boolean flag) {
        matchIfMetadataSilent = flag;
    }

    /**
     * Gets whether the attribute has been approved.
     * 
     * @return whether the attribute has been approved.
     */
    public boolean isOnlyIfChecked() {
        return onlyIfChecked;
    }

    /**
     * Sets whether the attribute has been approved.
     * 
     * @param flag whether the attribute has been approved.
     */
    public void setOnlyIfChecked(final boolean flag) {
        this.onlyIfChecked = flag;
    }

    /**
     * Sets the ServiceAccessDataModule.
     * 
     * @param serviceAccessDataModule The ServiceAccessDataModule to set.
     */
    public void setServiceAccessDataModule(final ServiceAccessDataModule serviceAccessDataModule) {
        this.serviceAccessDataModule = serviceAccessDataModule;
    }

    /**
     * Sets the AttributeReleaseModule.
     * 
     * @param attributeReleaseModule The AttributeReleaseModule to set.
     */
    public void setAttributeReleaseModule(final AttributeReleaseModule attributeReleaseModule) {
        this.attributeReleaseModule = attributeReleaseModule;
    }

    /** {@inheritDoc} */
    protected boolean doEvaluatePolicyRequirement(final ShibbolethFilteringContext filterContext)
            throws FilterProcessingException {
        throw new FilterProcessingException("This match functor is not supported in policy requirements");
    }

    /** {@inheritDoc} */
    protected boolean doEvaluateValue(final ShibbolethFilteringContext filterContext, final String attributeId,
            final Object attributeValue) throws FilterProcessingException {

        // TODO: this is a fair bit of work to do for every value, we might consider exposing
        // the "active" AttributeConsumingService object on the attribute request context.
        // I guess the intent was that filterContext.getAttributeRequestContext().getRequestedAttributesIds()
        // would be used. We could do that, but it wouldn't support value filtering, and I don't know how
        // easy it is for the request context code to get access to the the full set of attributes and
        // encoders from the resolver prior to resolution.

        // Check for SP role.
        final RoleDescriptor role = filterContext.getAttributeRequestContext().getPeerEntityRoleMetadata();
        if (!(role instanceof SPSSODescriptor)) {
            log.debug("attribute requester's metadata role does not contain attribute requirements");
            return false;
        }

        // If the request is an AuthnRequest, check for an AttributeConsumingServiceIndex.
        AttributeConsumingService service = null;
        final SAMLObject inbound = filterContext.getAttributeRequestContext().getInboundSAMLMessage();
        if (inbound != null && inbound instanceof AuthnRequest) {
            final Integer i = ((AuthnRequest) inbound).getAttributeConsumingServiceIndex();
            if (i != null) {
                final List<AttributeConsumingService> services =
                        ((SPSSODescriptor) role).getAttributeConsumingServices();
                for (final AttributeConsumingService s : services) {
                    if (s.getIndex() == i) {
                        service = s;
                        break;
                    }
                }
                if (service == null) {
                    log.warn("Incoming AuthnRequest's AttributeConsumingServiceIndex did not match peer's metadata");
                    return false;
                }
            }
        }

        final SAMLProfileRequestContext attributeRequestContext = filterContext.getAttributeRequestContext();
        log.debug("Context class is {}.", attributeRequestContext.getClass().getName());
        if (!(attributeRequestContext instanceof UApproveBaseSAMLProfileRequestContext)) {
            // In the context of the issuing response to SP or handling AttributeQuery profile, gets the approved
            // attributes stored in the storage.
            setReleasedAttributes(attributeRequestContext);
        }

        final boolean mandatoryMode = getMandatoryMode(attributeRequestContext);

        if (service == null) {
            service = ((SPSSODescriptor) role).getDefaultAttributeConsumingService();
        }
        if (service == null) {
            log.debug("The peer's metadata did not contain an AttributeConsumingService descriptor");
            if (matchIfMetadataSilent) {
                if (!doEvaluateOptionalAttribute(attributeId, mandatoryMode, attributeRequestContext)) {
                    return false;
                }
            }
            return matchIfMetadataSilent;
        }

        log.debug("Using AttributeConsumingService descriptor with index {}", service.getIndex());

        final BaseAttribute attr = filterContext.getUnfilteredAttributes().get(attributeId);
        if (attr != null && attr.getValues() != null) {
            final List<AttributeEncoder> encoders = attr.getEncoders();
            final Iterator<AttributeEncoder> i = encoders.iterator();
            while (i.hasNext()) {
                final AttributeEncoder encoder = i.next();
                final RequestedAttribute requested = UApprovejpSAMLHelper.findInMetadata(service, encoder);
                if (requested != null) {
                    if (onlyIfRequired && !requested.isRequired()) {
                        log.debug("Attribute {} requested in metadata, but was not required", attributeId);
                        return false;
                    }

                    if (!onlyIfRequired && !requested.isRequired()) {
                        if (!doEvaluateOptionalAttribute(attributeId, mandatoryMode, attributeRequestContext)) {
                            return false;
                        }
                    }

                    log.debug("Found attribute {} requested in metadata", attributeId);
                    final List<XMLObject> vals = requested.getAttributeValues();
                    if (vals == null || vals.isEmpty()) {
                        return true;
                    }
                    final String v = attributeValue.toString();
                    for (final XMLObject xmlObj : vals) {
                        if (match(xmlObj, v)) {
                            return true;
                        }
                    }
                    log.debug("Attribute {} found in metadata, but value not among those requested", attributeId);
                    return false;
                }
            }
            log.debug("Attribute {} not found in metadata", attributeId);
        }

        return false;
    }

    /**
     * Set released attributes from storage to the context.
     * 
     * In the context of the issuing response to SP or handling AttributeQuery profile, gets the approved attributes
     * stored in the storage.
     * 
     * @param attributeRequestContext
     */
    private void setReleasedAttributes(final SAMLProfileRequestContext attributeRequestContext) {
        log.debug("Get attributes release consent from storage.");
        Map<String, BaseAttribute> attributes = attributeRequestContext.getAttributes();
        if (attributes == null) {
            log.debug("attributeRequestContext.getAttributes() is null.");

            final String principalName = attributeRequestContext.getPrincipalName();
            final String relyingPartyId = attributeRequestContext.getPeerEntityId();

            final ConsentMode consentMode = serviceAccessDataModule.getConsentMode(principalName, relyingPartyId);
            if (consentMode == ConsentMode.GENERAL) {
                log.debug("User {} has general consented.", principalName);
                return;
            }

            final List<AttributeReleaseConsent> attributeReleaseConsents =
                    attributeReleaseModule.getAttributeReleaseConsents(principalName, relyingPartyId);

            attributes = new HashMap<String, BaseAttribute>();
            for (final AttributeReleaseConsent attr : attributeReleaseConsents) {
                final BasicAttribute<AttributeReleaseConsent> attribute = new BasicAttribute<AttributeReleaseConsent>();
                attribute.setId(attr.getAttributeId());
                final Collection<AttributeReleaseConsent> valuesHash = new ArrayList<AttributeReleaseConsent>();
                valuesHash.add(attr);
                attribute.setValues(valuesHash);
                log.debug("add attribute {}", attr.getAttributeId());
                attributes.put(attr.getAttributeId(), attribute);
            }
            attributeRequestContext.setAttributes(attributes);

        } else {
            log.debug("attributeRequestContext.getAttributes() is not null.");
        }
    }

    /**
     * Evaluates the status of consent of the attribute.
     * 
     * @param attributeId
     * @param mandatoryMode
     * @param requestedCtx TODO
     * @return TODO
     */
    protected boolean doEvaluateOptionalAttribute(final String attributeId, final boolean mandatoryMode,
            final AttributeRequestContext requestedCtx) {
        if (onlyIfChecked) {
            log.debug("Attribute {} is optional.", attributeId);
            if (mandatoryMode) {
                log.debug("(mandatory mode) attribute {} is optional.", attributeId);
                // In the mode of retrieving mandatory attributes, to treat as this attribute was not
                // approved.
                return false;
            }

            final Map<String, BaseAttribute> releasedAttributes = requestedCtx.getAttributes();
            if (releasedAttributes == null) {
                // Returns true in the context of the following:
                // (1) When displaying the attribute-release.jsp.
                // (2) When general consented.
                log.debug("skip compare with released attributes.");
            } else {
                log.debug("compare attribute to released attributes.");

                if (releasedAttributes.containsKey(attributeId)) {
                    log.debug("OK, this attribute is released. {}", attributeId);
                } else {
                    log.debug("This attribute is not released. {}", attributeId);
                    return false;
                }
            }
        } else {
            log.debug("\"onlyIfChecked\" is false. Behave same as saml:AttributeInMetadata.");
        }
        return true;
    }

    /**
     * @param attributeRequestContext TODO
     * @return
     */
    private boolean getMandatoryMode(final AttributeRequestContext attributeRequestContext) {
        if (!(attributeRequestContext instanceof UApproveBaseSAMLProfileRequestContext)) {
            return false;
        }
        if (((UApproveBaseSAMLProfileRequestContext) attributeRequestContext).isMandatoryMode()) {
            log.debug("filtering with mandatory mode.");
            return true;
        }
        return false;
    }

    /**
     * Checks whether an XMLObject's "value" matches a candidate value.
     * 
     * This method is imported from edu.internet2.middleware.shibboleth.common.attribute.filtering.provider.match.saml.
     * AttributeInMetadataMatchFunctor because it is a private member.
     * 
     * @param xmlObj the XMLObject to match
     * @param attributeValue the candidate value to match against
     * @return true if the two parameters are non-null and match
     */
    private boolean match(final XMLObject xmlObj, final String attributeValue) {
        // This is a substitute for a decoder layer that can generate
        // internal comparable value objects out of AttributeValue elements.
        // Short of that, some kind of pluggable comparison object with
        // knowledge of the XML syntax and the internal attribute objects
        // would be needed.
        String toMatch = null;
        if (xmlObj instanceof XSString) {
            toMatch = ((XSString) xmlObj).getValue();
        } else if (xmlObj instanceof XSURI) {
            toMatch = ((XSURI) xmlObj).getValue();
        } else if (xmlObj instanceof XSBoolean) {
            toMatch = ((XSBoolean) xmlObj).getValue().getValue() ? "1" : "0";
        } else if (xmlObj instanceof XSInteger) {
            toMatch = ((XSInteger) xmlObj).getValue().toString();
        } else if (xmlObj instanceof XSDateTime) {
            final DateTime dt = ((XSDateTime) xmlObj).getValue();
            if (dt != null) {
                toMatch = ((XSDateTime) xmlObj).getDateTimeFormatter().print(dt);
            }
        } else if (xmlObj instanceof XSBase64Binary) {
            toMatch = ((XSBase64Binary) xmlObj).getValue();
        } else if (xmlObj instanceof XSAny) {
            final XSAny wc = (XSAny) xmlObj;
            if (wc.getUnknownAttributes().isEmpty() && wc.getUnknownXMLObjects().isEmpty()) {
                toMatch = wc.getTextContent();
            }
        }
        if (toMatch != null) {
            return toMatch.equals(attributeValue);
        }
        log.warn("Unrecognized XMLObject type, unable to match as a string to candidate value");
        return false;
    }
}