/*
 * Copyright (c) 2014, National Institute of Informatics
 */

package jp.gakunin.shibboleth.idp.profile.saml1;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import jp.gakunin.uApprovejp.consent.ConsentMode;
import jp.gakunin.uApprovejp.sad.ServiceAccessData;
import jp.gakunin.uApprovejp.sad.ServiceAccessDataModule;

import org.apache.commons.lang.Validate;
import org.opensaml.saml1.core.AttributeStatement;
import org.opensaml.saml1.core.NameIdentifier;
import org.opensaml.saml1.core.Response;
import org.opensaml.saml1.core.Statement;
import org.opensaml.saml1.core.StatusCode;
import org.opensaml.ws.transport.http.HTTPInTransport;
import org.opensaml.ws.transport.http.HTTPOutTransport;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import ch.SWITCH.aai.uApprove.ar.AttributeReleaseConsent;
import ch.SWITCH.aai.uApprove.ar.AttributeReleaseHelper;
import ch.SWITCH.aai.uApprove.ar.AttributeReleaseModule;
import edu.internet2.middleware.shibboleth.common.attribute.BaseAttribute;
import edu.internet2.middleware.shibboleth.common.profile.ProfileException;
import edu.internet2.middleware.shibboleth.idp.profile.saml1.AttributeQueryProfileHandler;
import edu.internet2.middleware.shibboleth.idp.session.AuthenticationMethodInformation;
import edu.internet2.middleware.shibboleth.idp.session.Session;

/**
 * SAML 1 Approved Attribute Query profile handler.
 */
public class ApprovedAttributeQueryProfileHandler extends AttributeQueryProfileHandler {

    /** Class logger. */
    private final Logger log = LoggerFactory.getLogger(ApprovedAttributeQueryProfileHandler.class);

    private AttributeReleaseModule attributeReleaseModule;

    private ServiceAccessDataModule serviceAccessDataModule;

    /** Constructor. */
    public ApprovedAttributeQueryProfileHandler() {
        super();
        attributeReleaseModule = null;
        serviceAccessDataModule = null;
    }

    /**
     * Sets the AttributeReleaseModule.
     * 
     * @param attributeReleaseModule The AttributeReleaseModule to set.
     */
    public void setAttributeReleaseModule(final AttributeReleaseModule attributeReleaseModule) {
        this.attributeReleaseModule = attributeReleaseModule;
    }

    /**
     * Sets the ServiceAccessDataModule.
     * 
     * @param serviceAccessDataModule The ServiceAccessDataModule to set.
     */
    public void setServiceAccessDataModule(final ServiceAccessDataModule serviceAccessDataModule) {
        this.serviceAccessDataModule = serviceAccessDataModule;
    }

    /**
     * Initializes modules.
     */
    public void initialize() {
        log.debug("Initializing.");
        Validate.notNull(attributeReleaseModule, "attributeReleaseModule is not set.");
        Validate.notNull(serviceAccessDataModule, "serviceAccessDataModule is not set.");
    }

    /** {@inheritDoc} */
    public void processRequest(final HTTPInTransport inTransport, final HTTPOutTransport outTransport)
            throws ProfileException {
        final AttributeQueryUApproveContext requestContext = new AttributeQueryUApproveContext();
        Response samlResponse;
        try {
            decodeRequest(requestContext, inTransport, outTransport);

            if (requestContext.getProfileConfiguration() == null) {
                log.error("SAML 1 Attribute Query profile is not configured for relying party "
                        + requestContext.getInboundMessageIssuer());
                requestContext.setFailureStatus(buildStatus(
                        StatusCode.RESPONDER,
                        StatusCode.REQUEST_DENIED,
                        "SAML 1 Attribute Query profile is not configured for relying party "
                                + requestContext.getInboundMessageIssuer()));
                samlResponse = buildErrorResponse(requestContext);
            } else {
                resolvePrincipal(requestContext);

                final Session idpSession = getSessionManager().getSession(requestContext.getPrincipalName());
                if (idpSession != null) {
                    requestContext.setUserSession(idpSession);
                    final AuthenticationMethodInformation authnInfo =
                            idpSession.getAuthenticationMethods().get(requestContext.getInboundMessageIssuer());
                    if (authnInfo != null) {
                        requestContext.setPrincipalAuthenticationMethod(authnInfo.getAuthenticationMethod());
                    }
                }

                if (!requestContext.isTransient()) {
                    updateServiceAccessData(requestContext);
                }

                final ArrayList<Statement> statements = new ArrayList<Statement>();
                if (requestContext.getConsentMode() == ConsentMode.NOT_CONSENT) {
                    log.info("{} has not consented to {} .", requestContext.getPrincipalName(),
                            requestContext.getInboundMessageIssuer());
                } else if (!requestContext.isTransient() && requestContext.getConsentMode() == ConsentMode.ONE_TIME) {
                    log.info("{}'s current consent mode for {} is \"check always\".",
                            requestContext.getPrincipalName(), requestContext.getInboundMessageIssuer());
                } else {

                    resolveAttributes(requestContext);

                    final AttributeStatement attributeStatement =
                            buildAttributeStatement(requestContext, "urn:oasis:names:tc:SAML:1.0:cm:sender-vouches");
                    if (attributeStatement != null) {
                        requestContext.setReleasedAttributes(requestContext.getAttributes().keySet());
                        statements.add(attributeStatement);
                    }

                }

                samlResponse = buildResponse(requestContext, statements);
            }
        } catch (final ProfileException e) {
            samlResponse = buildErrorResponse(requestContext);
        }

        requestContext.setOutboundSAMLMessage(samlResponse);
        requestContext.setOutboundSAMLMessageId(samlResponse.getID());
        requestContext.setOutboundSAMLMessageIssueInstant(samlResponse.getIssueInstant());
        encodeResponse(requestContext);
        writeAuditLogEntry(requestContext);
    }

    /**
     * @param requestContext
     */
    private void updateServiceAccessData(final AttributeQueryUApproveContext requestContext) {
        final ServiceAccessData serviceAccessData = requestContext.getServiceAccessData();
        final ConsentMode consentMode = requestContext.getConsentMode();
        if (consentMode == ConsentMode.THIS_SERVICE || consentMode == ConsentMode.GENERAL) {
            serviceAccessData.updateBackChannelAccess();
            final String principalName = requestContext.getPrincipalName();
            serviceAccessDataModule.update(principalName, serviceAccessData);
        }
    }

    /**
     * Resolves the attributes for the principal.
     * 
     * @param requestContext current request context
     * 
     * @throws ProfileException thrown if there is a problem resolved attributes
     */
    protected void resolveAttributes(final AttributeQueryUApproveContext requestContext) throws ProfileException {
        super.resolveAttributes(requestContext);

        if (requestContext.isTransient() || requestContext.getConsentMode() != ConsentMode.THIS_SERVICE) {
            return;
        }

        final String principalName = requestContext.getPrincipalName();
        final String relyingPartyId = requestContext.getInboundMessageIssuer();
        final List<AttributeReleaseConsent> attributeReleaseConsents =
                attributeReleaseModule.getAttributeReleaseConsents(principalName, relyingPartyId);

        @SuppressWarnings("rawtypes") final Map<String, BaseAttribute> attributes = requestContext.getAttributes();
        AttributeReleaseHelper.resolveApprovedAttributes(attributes, attributeReleaseConsents,
                attributeReleaseModule.isCompareAttributeValues());
    }

    /** Basic data structure used to accumulate information as a request is being processed. */
    protected class AttributeQueryUApproveContext extends AttributeQueryContext {
        /** URI for transient name format. */
        private static final String TRANSIENT = "urn:mace:shibboleth:1.0:nameIdentifier";

        private ConsentMode consentMode;

        private String format;

        private ServiceAccessData serviceAccessData;

        /**
         * @return Returns the serviceAccessData.
         */
        public ServiceAccessData getServiceAccessData() {
            if (serviceAccessData == null) {
                final String principalName = getPrincipalName();
                if (principalName == null) {
                    // TODO
                }
                final String relyingPartyId = getInboundMessageIssuer();
                log.info("principalName={} relyingPartyId={}", principalName, relyingPartyId);
                serviceAccessData = serviceAccessDataModule.getServiceAccessData(principalName, relyingPartyId);
            }
            return serviceAccessData;
        }

        /**
         * @return
         * @throws ProfileException
         */
        public boolean isTransient() throws ProfileException {
            if (format == null) {
                format = getNameIDFormat();
                log.debug("NameIdentifier.getFormat()={}", format);
            }
            return format.equals(TRANSIENT);
        }

        /**
         * @return
         * @throws ProfileException
         */
        private String getNameIDFormat() throws ProfileException {
            final NameIdentifier nameID = getSubjectNameIdentifier();
            if (nameID == null) {
                setFailureStatus(buildStatus(StatusCode.RESPONDER, StatusCode.REQUEST_DENIED,
                        "Cannot get NameIdentifier"));
                final String msg = "Cannot get NameIdentifier from SAML request from relying party.";
                log.warn(msg);
                throw new ProfileException(msg);
            }
            return nameID.getFormat();
        }

        /**
         * @return
         */
        public ConsentMode getConsentMode() {
            if (consentMode != null) {
                return consentMode;
            }
            if (serviceAccessData == null) {
                getServiceAccessData();
            }
            consentMode = ConsentMode.getConsentMode(serviceAccessData);
            log.info("consentMode={}", consentMode);
            return consentMode;
        }
    }
}