/*
 * Copyright (c) 2014, National Institute of Informatics
 */

package jp.gakunin.uApprovejp.ar;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import javax.xml.namespace.QName;

import jp.gakunin.uApprovejp.lcsp.ServiceNames;
import jp.gakunin.uApprovejp.lcsp.ServiceNamesBuilder;
import jp.gakunin.uApprovejp.saml2.metadata.Description;
import jp.gakunin.uApprovejp.saml2.metadata.RequestedAttributeExtension;
import jp.gakunin.uApprovejp.saml2.metadata.UApprovejpMetadata;

import org.opensaml.saml2.common.Extensions;
import org.opensaml.saml2.core.Attribute;
import org.opensaml.saml2.metadata.AttributeConsumingService;
import org.opensaml.saml2.metadata.EntityDescriptor;
import org.opensaml.saml2.metadata.LocalizedString;
import org.opensaml.saml2.metadata.RequestedAttribute;
import org.opensaml.saml2.metadata.SPSSODescriptor;
import org.opensaml.saml2.metadata.ServiceName;
import org.opensaml.saml2.metadata.provider.MetadataProvider;
import org.opensaml.saml2.metadata.provider.MetadataProviderException;
import org.opensaml.samlext.saml2mdui.DisplayName;
import org.opensaml.samlext.saml2mdui.UIInfo;
import org.opensaml.xml.XMLObject;
import org.opensaml.xml.util.AttributeMap;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import ch.SWITCH.aai.uApprove.ar.SAMLHelper;
import edu.internet2.middleware.shibboleth.common.attribute.encoding.AttributeEncoder;
import edu.internet2.middleware.shibboleth.common.attribute.encoding.SAML1AttributeEncoder;
import edu.internet2.middleware.shibboleth.common.attribute.encoding.SAML2AttributeEncoder;

/**
 *
 */
public class UApprovejpSAMLHelper {

    /** Class logger. */
    private final Logger logger = LoggerFactory.getLogger(UApprovejpSAMLHelper.class);

    private final SAMLHelper samlHelper;

    /** The metadata provider. */
    private final MetadataProvider metadataProvider;

    /**
     * Constructor.
     * 
     * @param samlHelper
     */
    public UApprovejpSAMLHelper(final SAMLHelper samlHelper) {
        this.samlHelper = samlHelper;
        metadataProvider = samlHelper.getMetadataProvider();
    }

    /**
     * @param relyingPartyId
     * @param attributes
     */
    public void resolveDescriptions(final String relyingPartyId, final List<? extends AttributeDetail> attributes) {

        EntityDescriptor entityDescriptor = null;
        try {
            entityDescriptor = samlHelper.getEntityDescriptor(relyingPartyId);
            if (entityDescriptor == null) {
                logger.trace("No entity descriptor found for entityId {}.", relyingPartyId);
                // return new RelyingParty(relyingPartyId);
            }
        } catch (final MetadataProviderException e) {
            logger.error("Error while retrieving metadata descriptor for {}.", relyingPartyId, e);
            throw new IllegalStateException(e);
        }

        SPSSODescriptor spSSODescriptor = null;
        for (final String protocol : SAMLHelper.protocols) {
            spSSODescriptor = entityDescriptor.getSPSSODescriptor(protocol);
            if (spSSODescriptor != null) {
                logger.info("SPSSODescriptor found. protocol: {}.", protocol);
                break;
            }
            logger.info("{} does not support The protocol {}.", relyingPartyId, protocol);
        }

        if (spSSODescriptor == null) {
            logger.warn("No SPSSODescriptor available for relying party {}", relyingPartyId);
            // TODO
            return;
        }

        final AttributeConsumingService service = SAMLHelper.getAttributeConsumingService(spSSODescriptor);
        if (service == null) {
            logger.debug("No AttributeConsumingService available for relying party {}", relyingPartyId);
            return;
        }

        if (service.isDefault()) {
            logger.debug("Using DefaultAttributeConsumingService descriptor");
        } else {
            logger.debug("Using AttributeConsumingService descriptor with index {}", service.getIndex());
        }

        final Extensions extensions = spSSODescriptor.getExtensions();
        final List<RequestedAttributeExtension> requestedAttributeExtensions =
                getRequestedAttributeExtensions(extensions);
        final Map<String, List<Description>> descMap = new HashMap<String, List<Description>>();
        for (final RequestedAttributeExtension raExtension : requestedAttributeExtensions) {
            final String friendlyName = raExtension.getFriendlyName();
            logger.debug("RequestedAttributeExtension.FriendryName=" + friendlyName);

            for (final Description desc : raExtension.getDescriptions()) {
                final LocalizedString description = desc.getDescription();
                logger.debug("Description.lang=" + description.getLanguage());
                logger.debug("Description.desc=" + description.getLocalString());
            }

            descMap.put(friendlyName, raExtension.getDescriptions());
        }

        for (final AttributeDetail attribute : attributes) {
            final List<AttributeEncoder> encoders = attribute.getEncoders();
            // FIXME encorder and protocol

            final Iterator<AttributeEncoder> i = encoders.iterator();
            while (i.hasNext()) {
                final AttributeEncoder encoder = i.next();
                final RequestedAttribute requested = findInMetadata(service, encoder);

                if (requested != null) {
                    final String friendlyName = requested.getFriendlyName();
                    if (logger.isDebugEnabled()) {
                        // final QName schemaType = requested.getSchemaType();
                        final AttributeMap unknownAttrsMap = requested.getUnknownAttributes();
                        logger.debug("{} friendlyName={}", attribute.getId(), friendlyName);
                        for (final QName name : unknownAttrsMap.keySet()) {
                            logger.debug("    {} {}", name.getNamespaceURI(), name.toString());
                        }
                    }

                    final String attrDesc = requested.getUnknownAttributes().get(UApprovejpMetadata.ATTRDESC_NAME);
                    if (attrDesc != null) {
                        logger.debug("description: {}", attrDesc);
                        attribute.setIntendedUse(attrDesc);
                    } else {
                        logger.debug("no description");
                    }

                    if (descMap.containsKey(friendlyName)) {
                        logger.debug("{} has multilingal descriptions", friendlyName);
                        final List<Description> list = descMap.get(friendlyName);
                        attribute.setIntendedUseList(list);
                    } else {
                        logger.debug("{} has no descriptions", friendlyName);
                    }

                    // TODO solve conflicts
                    break;
                }
            }
        }
    }

    /**
     * @param service
     * @param encoder
     * @return
     */
    public static RequestedAttribute findInMetadata(final AttributeConsumingService service,
            final AttributeEncoder encoder) {
        // I think it would be cleaner to expose a "protocol" property on the
        // encoder interface to match up against the outbound protocol than to
        // do a switch here based on casting/detecting the encoder type based
        // on the protocol string and ending up with a maintenance issue, so
        // for now, I'm just looking for any matching Attribute name.
        // Still have to cast because of where the name qualifier is.
        // We could add an abstracted qualifier accessor on the encoder base
        // interface, and implement it via getNamespace / getNameFormat also.
        final List<RequestedAttribute> requested = service.getRequestAttributes();
        for (final RequestedAttribute attr : requested) {
            if (attr.getName().equals(encoder.getAttributeName())) {
                String qualifier = null;
                if (encoder instanceof SAML2AttributeEncoder) {
                    qualifier = ((SAML2AttributeEncoder) encoder).getNameFormat();
                } else if (encoder instanceof SAML1AttributeEncoder) {
                    qualifier = ((SAML1AttributeEncoder) encoder).getNamespace();
                }
                final String nameFormat = attr.getNameFormat();

                if (qualifier == null || nameFormat == null || nameFormat.equals(Attribute.UNSPECIFIED)
                        || qualifier.equals(nameFormat)) {
                    return attr;
                }
            }
        }
        return null;
    }

    /**
     * create RequestedAttributeExtension list from /EntitiesDescriptor/EntityDescriptor/SPSSODescriptor/Extensions
     * element.
     * 
     * RequestedAttributeExtension element belongs to the namespace "http://www.gakunin.jp/ns/uapprove-jp/metadata".
     */
    private List<RequestedAttributeExtension> getRequestedAttributeExtensions(final Extensions ext) {

        final List<RequestedAttributeExtension> raExtensionList = new ArrayList<RequestedAttributeExtension>();

        for (final XMLObject obj : ext.getOrderedChildren()) {
            logger.debug("Extensions/{}", obj.getElementQName());
            logger.debug("class={}", obj.getClass().getName());
            if (obj instanceof RequestedAttributeExtension) {
                raExtensionList.add((RequestedAttributeExtension) obj);
            }
        }
        return raExtensionList;
    }

    /**
     * Reads the relying party.
     * 
     * @param relyingPartyId The relying party id.
     * @param locale The locale.
     * @return Returns a relying party.
     */
    public ServiceNames readServiceNames(final String relyingPartyId) {
        final ServiceNamesBuilder builder = new ServiceNamesBuilder();
        builder.setRelyingPartyId(relyingPartyId);

        EntityDescriptor entityDescriptor = null;
        try {
            entityDescriptor = metadataProvider.getEntityDescriptor(relyingPartyId);
            if (entityDescriptor == null) {
                logger.trace("No entity descriptor found for entityId {}.", relyingPartyId);
                return builder.buildObject();
            }
        } catch (final MetadataProviderException e) {
            logger.error("Error while retrieving metadata descriptor for {}.", relyingPartyId, e);
            throw new IllegalStateException(e);
        }

        // (a) md:SPSSODescriptor/md:Extensions/mdui:UIInfo/mdui:DisplayName

        SPSSODescriptor spSSODescriptor = null;
        for (final String protocol : SAMLHelper.protocols) {
            spSSODescriptor = entityDescriptor.getSPSSODescriptor(protocol);
            if (spSSODescriptor != null) {
                logger.info("SPSSODescriptor found. protocol: {}.", protocol);
                break;
            }
            logger.info("{} does not support The protocol {}.", relyingPartyId, protocol);
        }

        if (spSSODescriptor == null) {
            logger.warn("No SPSSODescriptor available for relying party {}", relyingPartyId);
        }

        if (spSSODescriptor != null) {
            final Extensions ext = spSSODescriptor.getExtensions();
            if (ext != null) {
                final UIInfo uiinfo = getMduiUIInfo(ext);
                if (uiinfo != null) {
                    final List<DisplayName> displayNames = uiinfo.getDisplayNames();
                    builder.setDisplayNames(displayNames);
                }
            }

            // (b) md:AttributeConsumingService/md:ServiceName
            final AttributeConsumingService attrService = getAttributeConsumingService(spSSODescriptor);
            if (attrService == null) {
                logger.debug("No AttributeConsumingService available for relying party {}", relyingPartyId);
            } else {
                logger.trace("Relying party: {}, localized names {}, localized descriptions {}.", new Object[] {
                        relyingPartyId, attrService.getNames(), attrService.getDescriptions(),});

                final List<ServiceName> serviceNames = attrService.getNames();
                builder.setNames(serviceNames);
            }
        }

        return builder.buildObject();
    }

    /**
     * get mdui:UIInfo element from SPSSODescriptor/Extensions.
     * 
     * @param ext
     * @return UIInfo object, or null when mdui:UIInfo is not defined.
     */
    private UIInfo getMduiUIInfo(final Extensions ext) {
        logger.debug("get UIInfo from Extensions.");
        UIInfo uiinfo = null;
        for (final XMLObject object : ext.getOrderedChildren()) {
            if (object instanceof UIInfo) {
                uiinfo = (UIInfo) object;
                logger.debug("UIInfo#displayNames={}", uiinfo.getDisplayNames());
                break;
            }
        }
        return uiinfo;
    }

    /**
     * Gets the attribute consuming service.
     * 
     * @param spSSODescriptor The entity descriptor.
     * @return Returns the attribute consuming service, or null if undefined
     */
    protected AttributeConsumingService getAttributeConsumingService(final SPSSODescriptor spSSODescriptor) {
        final AttributeConsumingService result = spSSODescriptor.getDefaultAttributeConsumingService();
        if (result != null) {
            logger.trace("DefaultAttributeConsumingService found.");
            return result;
        }

        final List<AttributeConsumingService> list = spSSODescriptor.getAttributeConsumingServices();
        if (list != null && !list.isEmpty()) {
            logger.trace("AttributeConsumingService with index 0 found.");
            return list.get(0);
        }

        return result;
    }

}
