/*
 * Copyright (c) 2014, National Institute of Informatics
 */

package jp.gakunin.uApprovejp.consent;

import javax.servlet.http.HttpServletRequest;

import jp.gakunin.uApprovejp.sad.ServiceAccessData;
import ch.SWITCH.aai.uApprove.InvalidModeException;

/**
 *
 */
public enum ConsentMode {
    /**
     * This time I agree to send this data. I will check the data again at next login.
     */
    ONE_TIME {
        public ConsentOperation getOperation(final ConsentOperationFacrtory factory) {
            return factory.createOneTimeConsentOperation();
        }

        /** @return always true. */
        public boolean requiresConsent() {
            return true;
        }

        /** {@inheritDoc} */
        public String getStatusMessage(final String principalName) {
            final StringBuilder sb = new StringBuilder("User ");
            sb.append(principalName);
            sb.append(" given one time consent.");
            return sb.toString();
        }
    },
    /**
     * I agree that the data same as this time will be sent automatically to this service in the future.
     */
    THIS_SERVICE {
        public ConsentOperation getOperation(final ConsentOperationFacrtory factory) {
            return factory.createThisServiceConsentOperation();
        }

        /** @return always true. */
        public boolean requiresConsent() {
            return true;
        }

        /** {@inheritDoc} */
        public String getStatusMessage(final String principalName) {
            final StringBuilder sb = new StringBuilder("User ");
            sb.append(principalName);
            sb.append(" has given consent for this service.");
            return sb.toString();
        }
    },
    /**
     * I fully agree that in the future all my data will be released to the service provider as required by the service
     * that I will access.
     */
    GENERAL {
        public ConsentOperation getOperation(final ConsentOperationFacrtory factory) {
            return factory.createGeneralConsentOperation();
        }

        /** @return always false. */
        public boolean requiresConsent() {
            return false;
        }

        /** {@inheritDoc} */
        public String getStatusMessage(final String principalName) {
            final StringBuilder sb = new StringBuilder("User ");
            sb.append(principalName);
            sb.append(" has given gerneral consent.");
            return sb.toString();
        }
    },
    /**
     * Indicate that user has not consented.
     */
    NOT_CONSENT {
        public ConsentOperation getOperation(final ConsentOperationFacrtory factory) {
            throw new UnsupportedOperationException("");
        }

        /** @return always true. */
        public boolean requiresConsent() {
            return true;
        }

        /** {@inheritDoc} */
        public String getStatusMessage(final String principalName) {
            final StringBuilder sb = new StringBuilder("User ");
            sb.append(principalName);
            sb.append(" has not given consent not yet.");
            return sb.toString();
        }

    },
    ;

    public static final String Parameter = "ConsentMode";

    public static ConsentMode getConsentMode(final HttpServletRequest req) throws InvalidModeException {
        final String consentMode = req.getParameter(Parameter);

        try {
            return ConsentMode.valueOf(consentMode);
        } catch (final Exception e) {
            throw new InvalidModeException(consentMode, e);
        }
    }

    /**
     * @param serviceAccessData
     * @return
     */
    public static ConsentMode getConsentMode(final ServiceAccessData serviceAccessData) {
        if (!serviceAccessData.isConsented()) {
            return NOT_CONSENT;
        }
        if (serviceAccessData.isCheckAlways()) {
            return ONE_TIME;
        }
        if (serviceAccessData.isGeneralConsented()) {
            return GENERAL;
        }
        return THIS_SERVICE;
    }

    /**
     * Determine if consent is required.
     * 
     * @return Returns true if consent is required.
     */
    public abstract boolean requiresConsent();

    /**
     * @param factory
     * @return
     */
    public abstract ConsentOperation getOperation(ConsentOperationFacrtory factory);

    /**
     * Get a message about current consent status.
     * 
     * <p>
     * example: "User <var>princilalName</var> has given consent."
     * </p>
     * 
     * @param principalName
     * @return
     */
    public abstract String getStatusMessage(final String principalName);
}
