/*
 * Copyright (c) 2014, National Institute of Informatics
 */

package jp.gakunin.uApprovejp.lcsp;

import java.io.Serializable;
import java.io.UnsupportedEncodingException;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URLDecoder;
import java.net.URLEncoder;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;

import org.json.simple.JSONObject;
import org.opensaml.saml2.metadata.ServiceName;
import org.opensaml.samlext.saml2mdui.DisplayName;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * The service names for displaying on the page of approved SP list.
 */
public class ServiceNames implements Serializable {

    private static final long serialVersionUID = -2104044120672884532L;

    private final Logger logger = LoggerFactory.getLogger(ServiceNames.class);

    // JSON key
    private static final String UID = "UID";

    private static final String ENTITY_ID = "entityId";

    private static final String DISPLAY_NAMES = "displayNameMap";

    private static final String SERVICE_NAMES = "serviceNameMap";

    private static final int NAME_LENGTH_LIMIT = 60;

    private static final int COLUMN_LENGTH_LIMIT = 8192;

    private String entityId;

    private Map<String, String> displayNameMap;

    private Map<String, String> serviceNameMap;

    private ServiceNames(final String entityId) {
        this.entityId = entityId;
    }

    /**
     * @param relyingPartyId
     * @return
     */
    public static ServiceNames createInstance(final String relyingPartyId) {
        final ServiceNames obj = new ServiceNames(relyingPartyId);
        obj.displayNameMap = Collections.emptyMap();
        obj.serviceNameMap = Collections.emptyMap();
        return obj;
    }

    /**
     * Create an instance from a JSON object that is created by {@link #serialize()}.
     * 
     * @param serialized TODO
     * @return
     */
    public static ServiceNames createInstance(final JSONObject serialized) {
        final ServiceNames obj = new ServiceNames(null);
        obj.displayNameMap = new HashMap<String, String>();
        obj.serviceNameMap = new HashMap<String, String>();
        obj.desirialize(serialized);
        return obj;
    }

    /**
     * @return Returns the entityId.
     */
    public String getEntityId() {
        return entityId;
    }

    /**
     * @param displayNames
     */
    public void setMduiDisplayNames(final List<DisplayName> displayNames) {
        if (!displayNames.isEmpty()) {
            displayNameMap = new HashMap<String, String>();
            for (final DisplayName element : displayNames) {
                displayNameMap.put(element.getName().getLanguage(), element.getName().getLocalString());
            }
        }
    }

    /**
     * @param serviceNames
     */
    public void setServiceNames(final List<ServiceName> serviceNames) {
        if (!serviceNames.isEmpty()) {
            serviceNameMap = new HashMap<String, String>();
            for (final ServiceName element : serviceNames) {
                serviceNameMap.put(element.getName().getLanguage(), element.getName().getLocalString());
            }
        }
    }

    /**
     * Display the serviceName. Follow the definition of
     * {@link edu.internet2.middleware.shibboleth.idp.ui.ServiceNameTag} class, this is taken in order
     * <ol>
     * <li>SPSSODescriptor/Extensions/mdui:DisplayName (localized)</li>
     * <li>SPSSODescriptor/AttributeConsumeService/ServiceName (localized)</li>
     * <li>HostName from the Entity ID (when the scheme of the ID is "http" or "https")</li>
     * <li>Entity ID</li> </pre>
     * 
     * @param locale TODO
     * @return
     */
    public String getLocalizedServiceName(final Locale locale) {
        if (logger.isTraceEnabled()) {
            final StringBuilder sb = new StringBuilder();
            sb.append("getLocalizedServiceName(").append(locale).append(")");
            sb.append(" entityId=").append(entityId);
            logger.trace(sb.toString());
        }

        final String lang = locale.getLanguage();
        logger.debug("get mdui:DisplayName xml:lang=\"{}\"", lang);
        String displayName = displayNameMap.get(lang);
        if (displayName == null) {
            logger.debug("get md:ServiceName xml:lang=\"{}\"", lang);
            displayName = serviceNameMap.get(lang);
            if (displayName == null) {
                try {
                    logger.debug("get hostname from entityID.");
                    final URI uri = new URI(entityId);
                    final String scheme = uri.getScheme();
                    if (scheme.equals("http") || scheme.equals("https")) {
                        displayName = uri.getHost();
                    } else {
                        logger.debug("get entityID");
                        displayName = entityId;
                    }
                } catch (final URISyntaxException e) {
                    logger.warn("", e);
                    // e.printStackTrace();
                    logger.debug("get entitiyID");
                    displayName = entityId;
                }
            }
        }
        logger.debug("get mdui:DisplayName");
        return displayName;
    }

    /**
     * @return
     */
    @SuppressWarnings("unchecked")
    public JSONObject serialize() {
        logger.trace("Attempt to serialize data for {}", entityId);
        final JSONObject result = new JSONObject();

        logger.trace("Attempt to serialize display names {}", displayNameMap);
        final JSONObject displayNameMapSerialized = new JSONObject();
        for (final String key : displayNameMap.keySet()) {
            String value = displayNameMap.get(key);
            logger.trace("key={} length={}", key, value.length());
            if (value.length() > NAME_LENGTH_LIMIT) {
                logger.debug("{}: truncate the {} subsequent characters of mdui:DisplayName.", key, NAME_LENGTH_LIMIT);
                value = value.substring(0, NAME_LENGTH_LIMIT);
            }
            final String encode = encode(value);
            displayNameMapSerialized.put(key, encode);
        }
        logger.trace("Serialized display names {}", displayNameMapSerialized);

        logger.trace("Attempt to serialize names {}", serviceNameMap);
        final JSONObject serviceNameMapSerialized = new JSONObject();
        for (final String key : serviceNameMap.keySet()) {
            String value = serviceNameMap.get(key);
            logger.trace("key={} length={}", key, value.length());
            if (value.length() > NAME_LENGTH_LIMIT) {
                logger.debug("{}: truncate the {} subsequent characters of md:ServiceName.", key, NAME_LENGTH_LIMIT);
                value = value.substring(0, NAME_LENGTH_LIMIT);
            }
            final String encode = encode(value);
            serviceNameMapSerialized.put(key, encode);
        }
        logger.trace("Serialized names {}", serviceNameMapSerialized);

        result.put(UID, serialVersionUID);
        result.put(ENTITY_ID, entityId);
        result.put(DISPLAY_NAMES, displayNameMapSerialized);
        result.put(SERVICE_NAMES, serviceNameMapSerialized);

        final int length = result.toString().length();
        logger.trace("ServiceNames object serialized length={}", length);
        if (length > COLUMN_LENGTH_LIMIT) {
            logger.debug("Object size ({}) over {}. Remove service names.", length, COLUMN_LENGTH_LIMIT);
            result.put(DISPLAY_NAMES, new JSONObject());
            result.put(SERVICE_NAMES, new JSONObject());

            logger.trace("ServiceNames object serialized length={}", result.toString().length());
        }

        logger.trace("Serialized data '{}'", result);
        return result;
    }

    private void desirialize(final JSONObject serialized) {
        // TODO check UID
        logger.debug("UID={}", serialized.get(UID));

        entityId = (String) serialized.get(ENTITY_ID);

        final JSONObject displayNameMapSerialized = (JSONObject) serialized.get(DISPLAY_NAMES);
        for (final Object key : displayNameMapSerialized.keySet()) {
            final String value = (String) displayNameMapSerialized.get(key);
            final String decode = decode(value);
            displayNameMap.put((String) key, decode);
        }

        final JSONObject serviceNameMapSerialized = (JSONObject) serialized.get(SERVICE_NAMES);
        for (final Object key : serviceNameMapSerialized.keySet()) {
            final String value = (String) serviceNameMapSerialized.get(key);
            final String decode = decode(value);
            serviceNameMap.put((String) key, decode);
        }
    }

    /**
     * @param value
     * @return
     */
    private String encode(final String value) {
        String encode = null;
        try {
            encode = URLEncoder.encode(value, "UTF-8");
        } catch (final UnsupportedEncodingException e) {
            logger.warn("URL encode failed. return null.", e);
        }
        return encode;
    }

    /**
     * @param value
     * @return
     */
    private String decode(final String value) {
        String decode = null;
        try {
            decode = URLDecoder.decode(value, "UTF-8");
        } catch (final UnsupportedEncodingException e) {
            logger.warn("URL decode failed. return null.", e);
        }
        return decode;
    }
}
