/*
 * Copyright (c) 2014, National Institute of Informatics
 */

package jp.gakunin.uApprovejp.sad;

import java.util.ArrayList;
import java.util.List;
import java.util.Locale;

import jp.gakunin.uApprovejp.consent.ConsentMode;
import jp.gakunin.uApprovejp.sad.storage.Storage;

import org.joda.time.DateTime;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 *
 */
public class ServiceAccessDataModule {
    /** Class logger. */
    private final Logger logger = LoggerFactory.getLogger(ServiceAccessDataModule.class);

    /** Module enabled. */
    private final boolean enabled;

    /** Audit log enabled. */
    private final boolean auditLogEnabled;

    /** Storage. */
    private Storage storage;

    public ServiceAccessDataModule() {
        enabled = false;
        auditLogEnabled = false;
        storage = null;
    }

    /**
     * Sets the storage.
     * 
     * @param storage The storage to set.
     */
    public void setStorage(final Storage storage) {
        this.storage = storage;
    }

    /**
     * Get service access data about this principal.
     * 
     * Check the following order:
     * <ol>
     * <li>general consent</li>
     * <li>consent of each service</li>
     * </ol>
     * 
     * @param principalName
     * @param relyingPartyId
     * @return
     */
    public ServiceAccessData getServiceAccessData(final String principalName, final String relyingPartyId) {
        String entityID = ServiceAccessData.WILDCARD;

        if (!storage.containsServiceAccessData(principalName, entityID)) {
            entityID = relyingPartyId;
            if (!storage.containsServiceAccessData(principalName, entityID)) {
                // The first access (or after reset).
                return new ServiceAccessData(relyingPartyId);
            }
        }

        return storage.readServiceAccessData(principalName, entityID);
    }

    /**
     * @param principalName
     * @param locale TODO
     * @return
     */
    public List<LocalizedServiceAccessData> getServiceAccessDatas(final String principalName, final Locale locale) {
        final List<ServiceAccessData> serviceAccessDatas = storage.readServiceAccessDatas(principalName);
        final List<LocalizedServiceAccessData> result = new ArrayList<LocalizedServiceAccessData>();
        for (final ServiceAccessData sad : serviceAccessDatas) {
            final LocalizedServiceAccessData localizedServiceAccessData = new LocalizedServiceAccessData(sad, locale);
            result.add(localizedServiceAccessData);
        }
        return result;
    }

    /**
     * Clear service access data.
     * 
     * @param principalName
     * @param relyingPartyId
     */
    public void clearServiceAccessData(final String principalName, final String relyingPartyId) {
        logger.info("Clear service access data for {}.", principalName);
        storage.deleteServiceAccessData(principalName, ServiceAccessData.WILDCARD);
        storage.deleteServiceAccessData(principalName, relyingPartyId);
    }

    /**
     * @param principalName
     * @param relyingPartyId
     */
    public void clearRelyingPartyConsent(final String principalName, final String relyingPartyId) {
        logger.info("Clear service access data for {}.", principalName);
        storage.deleteServiceAccessData(principalName, relyingPartyId);
    }

    /**
     * @param principalName
     */
    public void clearGeneralConsent(final String principalName) {
        logger.info("Clear service access data for {}.", principalName);
        storage.deleteServiceAccessData(principalName, ServiceAccessData.WILDCARD);
    }

    /**
     * 
     * @param principalName
     * @param relyingPartyId
     * @return
     */
    public ConsentMode getConsentMode(final String principalName, final String relyingPartyId) {
        final ServiceAccessData serviceAccessData = getServiceAccessData(principalName, relyingPartyId);
        final ConsentMode consentMode = ConsentMode.getConsentMode(serviceAccessData);

        logger.debug(consentMode.getStatusMessage(principalName));

        return consentMode;
    }

    /**
     * Create or update the consent information.
     * 
     * @param principalName
     * @param relyingPartyId
     * @param serializedServiceName TODO
     * @param consentDate
     */
    public void consent(final String principalName, final String relyingPartyId, final String serializedServiceName,
            final DateTime consentDate) {

        ServiceAccessData serviceAccessData = storage.readServiceAccessData(principalName, relyingPartyId);
        if (serviceAccessData == null) {
            serviceAccessData = new ServiceAccessData(relyingPartyId, false, serializedServiceName, consentDate);
            storage.createServiceAccessData(principalName, serviceAccessData);
        } else {
            storage.updateServiceAccessData(principalName, relyingPartyId, serializedServiceName, consentDate, false);
        }
    }

    /**
     * Create or update the general consent information.
     * 
     * @param principalName
     * @param consentDate
     */
    public void consentGeneral(final String principalName, final DateTime consentDate) {
        storage.clearServiceAccessData(principalName);
        final ServiceAccessData serviceAccessData =
                new ServiceAccessData(ServiceAccessData.WILDCARD, false, null, consentDate);
        storage.createServiceAccessData(principalName, serviceAccessData);
    }

    /**
     * Create or update the One-Time consent information.
     * 
     * @param principalName
     * @param relyingPartyId
     * @param consentDate
     */
    public void consentOnetime(final String principalName, final String relyingPartyId, final DateTime consentDate) {

        if (!storage.containsServiceAccessData(principalName, relyingPartyId)) {
            final ServiceAccessData serviceAccessData = new ServiceAccessData(relyingPartyId, true, null, consentDate);
            storage.createServiceAccessData(principalName, serviceAccessData);
        } else {
            storage.updateServiceAccessData(principalName, relyingPartyId, null, consentDate, true);
        }
    }

    /**
     * @param principalName
     * @param serviceAccessData
     */
    public void update(final String principalName, final ServiceAccessData serviceAccessData) {
        final String relyingPartyId = serviceAccessData.getRelyingPartyId();
        storage.updateSADBackChannelAccess(principalName, relyingPartyId, serviceAccessData);
    }
}
