/*
 * Copyright (c) 2014, National Institute of Informatics
 */

package jp.gakunin.uApprovejp.sad.storage;

import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Timestamp;
import java.util.Collections;
import java.util.Date;
import java.util.List;

import jp.gakunin.uApprovejp.sad.ServiceAccessData;

import org.joda.time.DateTime;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.dao.DataAccessException;
import org.springframework.dao.EmptyResultDataAccessException;
import org.springframework.jdbc.core.simple.ParameterizedRowMapper;

import ch.SWITCH.aai.uApprove.AbstractJDBCStorage;

/** JDBC implementation. */
public class JDBCStorage extends AbstractJDBCStorage implements Storage {

    /** Class logger. */
    @SuppressWarnings("unused")
    private final Logger logger = LoggerFactory.getLogger(JDBCStorage.class);

    /** {@see ServiceAccessData} row mapper. */
    private static final class ServiceAccessDataMapper implements ParameterizedRowMapper<ServiceAccessData> {
        /** {@inheritDoc} */
        public ServiceAccessData mapRow(final ResultSet rs, final int rowNum) throws SQLException {
            final String relyingPartyId = rs.getString("relyingPartyId");
            final boolean checkAlways = rs.getBoolean("checkAlways");
            final String serviceNames = rs.getString("serviceNames");
            final DateTime consentDateDT = new DateTime(rs.getTimestamp("consentDate"));
            final int backChannelAccessCount = rs.getInt("backChannelAccessCount");
            final DateTime backChannelLastAccessDT;
            if (backChannelAccessCount > 0) {
                final Timestamp backChannelLastAccess = rs.getTimestamp("backChannelLastAccess");
                backChannelLastAccessDT = new DateTime(backChannelLastAccess);
            } else {
                backChannelLastAccessDT = null;
            }

            final ServiceAccessData serviceAccessData =
                    new ServiceAccessData(relyingPartyId, checkAlways, serviceNames, consentDateDT,
                            backChannelLastAccessDT, backChannelAccessCount);
            return serviceAccessData;
        }
    }

    private final ServiceAccessDataMapper serviceAccessDataMapper = new ServiceAccessDataMapper();

    /** Default constructor. */
    public JDBCStorage() {
        super();
    }

    /** {@inheritDoc} */
    public void createServiceAccessData(final String userId, final ServiceAccessData serviceAccessData) {
        try {
            final Date consentDate = serviceAccessData.getConsentDate().toDate();

            getJdbcTemplate().update(getSqlStatements().getProperty("createServiceAccessData"), userId,
                    serviceAccessData.getRelyingPartyId(), serviceAccessData.isCheckAlways(),
                    serviceAccessData.getServiceNames(), consentDate);
        } catch (final DataAccessException e) {
            handleDataAccessException(e);
        }

    }

    /** {@inheritDoc} */
    public ServiceAccessData readServiceAccessData(final String userId, final String relyingPartyId) {
        try {
            return getJdbcTemplate().queryForObject(getSqlStatements().getProperty("readServiceAccessData"),
                    serviceAccessDataMapper, userId, relyingPartyId);
        } catch (final EmptyResultDataAccessException e) {
            return null;
        } catch (final DataAccessException e) {
            handleDataAccessException(e);
            return null;
        }
    }

    /** {@inheritDoc} */
    public void updateServiceAccessData(final String userId, final String relyingPartyId, final String serviceNames,
            final DateTime consentDate, final boolean onetime) {
        try {
            if (onetime) {
                getJdbcTemplate().update(getSqlStatements().getProperty("resetServiceAccessData"), onetime,
                        serviceNames, consentDate.toDate(), userId, relyingPartyId);
            } else {
                getJdbcTemplate().update(getSqlStatements().getProperty("updateServiceAccessData"), onetime,
                        serviceNames, consentDate.toDate(), userId, relyingPartyId);
            }
        } catch (final DataAccessException e) {
            handleDataAccessException(e);
        }
    }

    /** {@inheritDoc} */
    public void deleteServiceAccessData(final String userId, final String relyingPartyId) {
        try {
            getJdbcTemplate().update(getSqlStatements().getProperty("deleteServiceAccessData"), userId, relyingPartyId);
        } catch (final DataAccessException e) {
            handleDataAccessException(e);
        }
    }

    /** {@inheritDoc} */
    public void clearServiceAccessData(final String userId) {
        try {
            getJdbcTemplate().update(getSqlStatements().getProperty("clearServiceAccessDatas"), userId);
        } catch (final DataAccessException e) {
            handleDataAccessException(e);
        }
    }

    /** {@inheritDoc} */
    public boolean containsServiceAccessData(final String userId, final String relyingPartyId) {
        try {
            return getJdbcTemplate().queryForInt(getSqlStatements().getProperty("containsServiceAccessData"), userId,
                    relyingPartyId) > 0;
        } catch (final DataAccessException e) {
            handleDataAccessException(e);
            return false;
        }
    }

    /** {@inheritDoc} */
    public void updateSADBackChannelAccess(final String userId, final String relyingPartyId,
            final ServiceAccessData serviceAccessData) {
        try {
            final Date lastAccess = serviceAccessData.getBackChannelLastAccess().toDate();
            getJdbcTemplate().update(getSqlStatements().getProperty("updateSADBackChannelAccess"), lastAccess,
                    serviceAccessData.getBackChannelAccessCount(), userId, relyingPartyId);
        } catch (final DataAccessException e) {
            handleDataAccessException(e);
        }
    }

    /** {@inheritDoc} */
    public List<ServiceAccessData> readServiceAccessDatas(final String userId) {
        try {
            return getJdbcTemplate().query(getSqlStatements().getProperty("readServiceAccessDatas"),
                    serviceAccessDataMapper, userId);
        } catch (final EmptyResultDataAccessException e) {
            return Collections.emptyList();
        } catch (final DataAccessException e) {
            handleDataAccessException(e);
            return Collections.emptyList();
        }
    }
}
